#!/usr/bin/env python
# coding: utf-8

from __future__ import print_function

from scipy.interpolate import interp1d

import itertools
import sys

from matplotlib.colors import LinearSegmentedColormap, LogNorm
import matplotlib.pyplot as plt
import numpy as np
import netCDF4
import shapely
import rasterio.features

import cartopy
import cartopy.crs as ccrs
import cartopy.feature as cfeature

import orbitsim as sim

#from . import orbit
from . import config as cfg


BERLIN = 13.407696, 52.516984

POWER_STATION = {
    'Jänschwalde':    (14.457934, 51.836579),
    'Schwarze Pumpe': (14+21/60+12/3600, 51+32/60+10/3600),
    'Boxberg':        (14+33/60+53/3600, 51+24/60+58/3600),
    'Lippendorf':     (12+22/60+40/3600, 51+11/60+ 7/3600),
    'Turów':          (14.913, 50.948),
    'Patnów':         (18.236118, 52.301086)
}

# with 0.05 border on Smartcarb domain
XLIM = (-1.0554999649524688, 5.2455001413822178)
YLIM = (2.7944999933242798, 8.1954997777938843)



def smartcarb_projection():
    return ccrs.RotatedPole(
        pole_latitude=cfg.pollat,
        pole_longitude=cfg.pollon
    )


def create_map(domain, add_colorbar=False, edgecolor='black'):
    if domain in['berlin', 'smartcarb']:
        projection = ccrs.RotatedPole(
            pole_latitude=cfg.pollat,
            pole_longitude=cfg.pollon
        )
    else:
        projection = ccrs.PlateCarree()

    if domain == 'smartcarb':
        bottom, top = cfg.startlat, cfg.stoplat
        left, right = cfg.startlon, cfg.stoplon

    elif domain == 'europe':
        left, right = -12, 42
        top, bottom = 72, 33

    elif domain == 'berlin':
        left, right = 1.8, 2.4
        bottom, top = 5.35,5.75

    else:
        left, right = -180, 180
        top, bottom = 80, -80

    dx = (right - left)
    dy = (top - bottom)

    if add_colorbar:
        dc = 1.0
        ll, rr = 0.017, 0.823
    else:
        dc = 0.0
        ll, rr = 0.02, 0.96

    fig = plt.figure(figsize=(6.0+dc, 6.0*dy/dx))
    ax = fig.add_axes([ll,0.02,rr,0.96], projection=projection)
    ax.set_xlim(left, right)
    ax.set_ylim(bottom, top)

    if add_colorbar:
        cax = fig.add_axes([2*ll+rr,0.02,0.04,0.96])
    else:
        cax = None

    ax.coastlines(resolution='10m', color=edgecolor, linewidth=1.0)
    lines = cfeature.NaturalEarthFeature(
        category='cultural',
        name='admin_0_boundary_lines_land',
        scale='10m',
    )
    ax.add_feature(lines, edgecolor=edgecolor, facecolor='none', linewidth=1.0)

    if domain in ['smartcarb', 'berlin']:
        lines = cfeature.NaturalEarthFeature(
            category='cultural',
            name='admin_1_states_provinces_lines',
            scale='10m',
        )
        ax.add_feature(lines, edgecolor=edgecolor, facecolor='none', linewidth=0.5)

    return fig, ax, cax


def add_city_labels(ax, color='black', size='large'):
    ax.plot(BERLIN[0], BERLIN[1], '.', color=color, transform=ccrs.PlateCarree())
    ax.text(BERLIN[0], BERLIN[1], 'Berlin ', va='center', ha='right', fontsize=size,
            transform=ccrs.PlateCarree(), color=color)

    for key, (lon, lat) in POWER_STATION.items():
        ax.plot(lon, lat, '.', color=color, transform=ccrs.PlateCarree())

        if key in ['Patnów', 'Boxberg']:
            ax.text(lon, lat, u'%s  ' % key, va='center', ha='right', fontsize=size,
                    transform=ccrs.PlateCarree(), color=color)
        else:
            ax.text(lon, lat, u'  %s' % key, va='center', fontsize=size,
                    transform=ccrs.PlateCarree(), color=color)


    #ax.plot(JANSCHWALDE[0], JANSCHWALDE[1], '.', color=color, transform=ccrs.PlateCarree())
    #ax.text(JANSCHWALDE[0], JANSCHWALDE[1], u'  Jänschwalde', va='center', fontsize=size,
    #        transform=ccrs.PlateCarree(), color=color)



def draw_smartcarb_domain(ax, color='black'):
    projection = ccrs.RotatedPole(
        pole_latitude=cfg.pollat,
        pole_longitude=cfg.pollon
    )
    lon = [cfg.startlon, cfg.startlon, cfg.stoplon, cfg.stoplon, cfg.startlon]
    lat = [cfg.startlat, cfg.stoplat, cfg.stoplat, cfg.startlat, cfg.startlat]

    ax.plot(lon, lat, '-', color=color, transform=projection)



def create_orbits(scenario_filename, start=0, length=1, nrow=None):
    """\
    Create list of orbit polgons for instrument scenario. `nrow` is the
    number of across-track pixels. A small number (nrow = 2) will be fast
    but might fail for wide swaths because the viewing angles for the corner
    points might be larger than 90 degree.
    """
    # init orbit simulator
    s = sim.Orbit_sim.from_file(scenario_filename)
    s.tsamp = 10.0
    s.nrow = s.nrow if nrow is None else nrow

    orbits = []
    for i in range(start, start+length):
        s.phase_start = i + 0.33#0.29
        s.phase_end = i + 0.37#0.71
        s.calculate_revolution()

        orbit = create_orbit_polygon(s)
        orbits.append(orbit)

    return orbits




def create_orbit_polygon(s):
    """\
    Create a polygon from the boundary of orbit (using center points).
    """
    lon, lat, lon_center, lat_center = orbit.reshape_orbit(s)
    nrows, ncols = lon.shape

    x = np.concatenate([lon[0,:], lon[:,-1], lon[-1,::-1], lon[::-1,0]])[::-1]
    y = np.concatenate([lat[0,:], lat[:,-1], lat[-1,::-1], lat[::-1,0]])[::-1]
    coords = np.transpose([x,y])
    orbit_poly = shapely.geometry.Polygon(coords)

    if not orbit_poly.is_valid or lon[-1,ncols//2] > lon[0,ncols//2]:
        lon[lon < 0] = 360.0 + lon[lon < 0]
        #lon[lon > 180] = lon[lon > 180] - 360.0
        x = np.concatenate([lon[0,:], lon[:,-1], lon[-1,::-1], lon[::-1,0]])[::-1]
        y = np.concatenate([lat[0,:], lat[:,-1], lat[-1,::-1], lat[::-1,0]])[::-1]
        coords = np.transpose([x,y])
        orbit_poly = shapely.geometry.Polygon(coords)

        if not orbit_poly.is_valid:
            print('Orbit not valid')

    return orbit_poly


def draw_polygon(polygon, ax, edgecolor='none', fillcolor='silver', alpha=0.5,
        transform=None, zorder=-1):
    """\
    Draw an orbit polygon on matplotlib axes.
    """
    if transform is None:
        transform = ccrs.PlateCarree()

    feature = cartopy.feature.ShapelyFeature([polygon], transform)
    ax.add_feature(feature, color=fillcolor, alpha=alpha, edgecolor=edgecolor, zorder=zorder)
    return feature


def rasterize_polygon(polygon):
    shape = (int(180/resolution), int(360/resolution))
    transform = rasterio.Affine(resolution,0,-180,0,resolution,-90)
    rasterio.features.rasterize([polygon], shape, transform=transform)


def rasterize(orbits, resolution=0.1):
    """\
    Rasterize orbits on lon/lat grid and returns array with number of overlaps.
    """
    # lon = [-180,180]
    shape = (int(180/resolution), int(360/resolution))
    transform = rasterio.Affine(resolution,0,-180,0,resolution,-90)
    img = np.array([rasterio.features.rasterize([orbit], shape, transform=transform)
         for orbit in orbits])

    # lon = [180,360]
    transform = rasterio.Affine(resolution,0,180,0,resolution,-90)
    img2 = np.array([rasterio.features.rasterize([orbit], shape, transform=transform)
        for orbit in orbits])

    return (img+img2).sum(0)


def calculate_pixel(scenario_filename, orbit_number):
    s = sim.Orbit_sim.from_file(scenario_filename)
    s.phase_start = orbit_number + 0.349
    s.phase_end = orbit_number + 0.354
    s.calculate_revolution()
    return orbit.reshape_orbit(s)


def make_map(scenario_filename, domain, orbit=0, length=14, nrow=11,
        add_labels=False, add_domain=False, add_pixels=False, do_overlap=False,
        vmin=None, vmax=None, size='large', color='white',
        orbits=None):

    if orbits is None:
        orbits = create_orbits(scenario_filename, orbit, length, nrow=nrow)

    fig, ax, cax = create_map(domain, add_colorbar=do_overlap)

    if add_labels:
        add_city_labels(ax, color=color)

    if add_domain:
        draw_smartcarb_domain(ax, color=color)

    if add_pixels:
        s = sim.Orbit_sim.from_file(scenario_filename)
        s.phase_start = orbit + 0.351
        s.phase_end = orbit + 0.354
        #s.nrow = 2
        s.calculate_revolution()
        lon, lat, lon_center, lat_center = sim.reshape_orbit(s)
        print(lon.shape, lon_center.shape)

        m,n = lon.shape

        for i in range(m):
            ax.plot(lon_center[i,:,0], lat_center[i,:,0], 'k-',
                    transform=ccrs.PlateCarree(),
                    linewidth=0.5, color='blue'
            )
        for j in range(n):
            ax.plot(lon_center[:,j,0], lat_center[:,j,0], 'k-',
                    transform=ccrs.PlateCarree(),
                    linewidth=0.5, color='blue'
            )

    if do_overlap:
        raster = rasterize(orbits, resolution=0.1)
        vmin = np.min(raster) if vmin is None else vmin
        vmax = np.max(raster) if vmax is None else vmax

        cmap = plt.cm.get_cmap('viridis', vmax-vmin+1)
        c = ax.imshow(raster, vmin=vmin-0.5, vmax=vmax+0.5, cmap=cmap,
                extent=[-180,180,-90,90], alpha=0.9)
        plt.colorbar(c, cax=cax).set_label('number of overlapping swaths', fontsize=size)
    else:
        for orbit in orbits:
            pass#draw_orbit(orbit, ax)





def load_cosmo_data(filename, tracer):

    with netCDF4.Dataset(filename, 'r') as nc:
        rlon = nc.variables['rlon'][:]
        rlat = nc.variables['rlat'][:]

        if tracer == 'XCO2':
            field1 = nc.variables['XCO2_TOT'][0]
            field2 = nc.variables['XCO2_GPP'][0]
            field3 = nc.variables['XCO2_RA'][0]
            field4 = nc.variables['XCO2_BG'][0]

            field = field1 - field2 + field3 + field4

        elif tracer == 'NO2':
            field = nc.variables['NO2_TOT_column'][0] + nc.variables['NO2_BG_column'][0]

        elif tracer == 'CO':
            field = nc.variables['CO_TOT_column'][0] + nc.variables['CO_BG_column'][0]

        else:
            field = nc.variables[tracer][0]

    return rlon,rlat,field


def make_cosmo_map(rlon, rlat, field, vmin=0.0, vmax=None, cmap='viridis',
        border=0.00, norm='linear', label='', domain='smartcarb', fig=None,
        clct=None, xlim=None, ylim=None, edgecolor='black',
        add_colorbar=True):

    dlon = rlon[1] - rlon[0]
    left = rlon[0] - dlon / 2.0
    right = rlon[-1] + dlon / 2.0

    dlat = rlat[1] - rlat[0]
    bottom = rlat[0] - dlat / 2.0
    top = rlat[-1] + dlat / 2.0

    if fig is None:
        fig,ax,cax = create_map(domain, add_colorbar, edgecolor=edgecolor)
    else:
        ax, cax = fig.get_axes()
        cax.cla()

        # remove cosmo data
        ax.axes.get_images()[0].remove()

    if norm == 'log':
        norm = LogNorm(vmin=vmin, vmax=vmax)
        c = ax.imshow(field, norm=norm, cmap=cmap, zorder=-1,
                extent=(left, right, bottom, top))
    else:
        c = ax.imshow(field, vmin=vmin, vmax=vmax, cmap=cmap, zorder=-1,
                extent=(left, right, bottom, top))

    if add_colorbar:
        cb = plt.colorbar(c, cax=cax)
        cb.set_label(label)

    if clct is not None:
        cm = LinearSegmentedColormap.from_list('alpha', [(1,1,1,0),(1,1,1,1)], 256)
        ax.imshow(clct, vmin=0, vmax=1, cmap=cm, zorder=-1,
                extent=(left, right, bottom, top))


    if domain == 'smartcarb':
        dlon = rlon[-1] - rlon[0]
        dlat = rlat[-1] - rlat[0]
        ax.set_xlim(left + border * dlon, right - border * dlon)
        ax.set_ylim(bottom + border * dlat, top - border * dlat)

    if xlim is not None:
        ax.set_xlim(*xlim)

    if ylim is not None:
        ax.set_ylim(*ylim)


    return fig



def load_orbit_data(filename, tracer):
    with netCDF4.Dataset(filename) as nc:
        lon = nc.variables['longitude_corners'][:]
        lat = nc.variables['latitude_corners'][:]

        if tracer == 'XCO2':
            field1 = nc.variables['XCO2_TOT'][:]
            field2 = nc.variables['XCO2_GPP'][:]
            field3 = nc.variables['XCO2_RA'][:]
            field4 = nc.variables['XCO2_BG'][:]

            field = field1 - field2 + field3 + field4

        elif tracer == 'NO2':
            field = nc.variables['NO2_TOT_column'][:] + nc.variables['NO2_BG_column'][:]

        elif tracer == 'CO':
            field = nc.variables['CO_TOT_column'][:] + nc.variables['CO_BG_column'][:]

        else:
            field = nc.variables[tracer][:]

    return lon, lat, field


def make_level2_map(lon, lat, values, fig=None, vmin=None, vmax=None,
        domain='smartcarb', label='', alpha=1.0, cmap='viridis',
        clct=None, xlim=XLIM, ylim=YLIM, edgecolor='black'):
    """\
    lon, lat :: longitude and latitude of pixel corners
    values   :: 2d field of level-2 data
    """
    if fig is None:
        fig,ax,cax = create_map(domain, True, edgecolor=edgecolor)
    else:
        ax, cax = fig.get_axes()

    # draw silver background
    ax.imshow(np.tile(np.array([[[188, 188, 188]]],
            dtype=np.uint8), [2, 2, 1]),
            origin='upper',
            transform=ccrs.PlateCarree(),
            extent=[-180, 180, -180, 180]
    )

    # create tiled grid
    if np.ndim(lon) == 3 and np.ndim(lat) == 3:
        lonc, latc = corners2grid(lon,lat)

        c = ax.pcolormesh(lonc, latc, values, vmin=vmin, vmax=vmax, cmap=cmap,
                alpha=alpha, transform=ccrs.PlateCarree())

        if clct is not None:
            cm = LinearSegmentedColormap.from_list('alpha', [(1,1,1,0),(1,1,1,1)], 256)
            ax.pcolormesh(lonc, latc, clct, vmin=0, vmax=1, cmap=cm,
                    transform=ccrs.PlateCarree())

    else:
        c = ax.scatter(lon, lat, c=values, vmin=vmin, vmax=vmax, cmap=cmap,
                alpha=alpha, transform=ccrs.PlateCarree())


    if xlim is not None:
        ax.set_xlim(*xlim)
    if ylim is not None:
        ax.set_ylim(*ylim)

    cb = plt.colorbar(c, cax=cax)
    cb.set_label(label)

    return fig



def corners2grid(lon, lat):
    lonc = np.zeros((lon.shape[0], lon.shape[1]+1))
    lonc[:,:-1] = lon[:,:,0]
    lonc[:,-1] = lon[:,-1,1]

    latc = np.zeros((lat.shape[0], lat.shape[1]+1))
    latc[:,:-1] = lat[:,:,0]
    latc[:,-1] = lat[:,-1,1]

    return lonc, latc





def calculate_xco2(xm, p, ps, h):

    p = p[:,::-1,:,:]
    h = p[:,::-1,:,:]
    xm = p[:,::-1,:,:]

    _, ke, je, ie = xm.shape

    xco2 = np.zeros((je,ie))

    for i,j in itertools.product(range(ie), range(je)):
        h_mid = 0.5 * (h[0,1:,j,i] + h[0,:-1,j,i])
        pp = np.interp(np.log(h[0,:,j,i]), np.log(h_mid), p[0,:,j,i])
        pp[0] = ps[0,j,i]

        mair = - np.diff(pp) / g
        xco2[j,i] = np.sum(mair * xm[0,j,i]) / np.sum(mair)

    return xco2





def main(field, rlon, rlat, rotpole=None, vmin=None, vmax=None, cmap=None,
        label=None, edgecolor='white', border=50, ax=None, img=None,
        title=None):
    """\
    Make map. `ax` and `img` can be already drawn map of which only
    the image is replaced with a new one (assuming same vmin, vmax etc.).
    """

    if label is None:
        label = 'no title'

    dlon = (rlon[1]-rlon[0])
    left = rlon[0] - dlon / 2.0
    right = rlon[-1] + dlon / 2.0

    dlat = (rlat[1]-rlat[0])
    bottom = rlat[0] - dlat / 2.0
    top = rlat[-1] + dlat / 2.0

    make_new = (ax is None and img is None)

    if make_new:

        if rotpole is None:
            rotated_pole = ccrs.RotatedPole(
                pole_latitude=cfg.pollat,
                pole_longitude=cfg.pollon
            )
        else:
            rotated_pole = ccrs.RotatedPole(
                pole_latitude=rotpole['grid_north_pole_latitude'],
                pole_longitude=rotpole['grid_north_pole_longitude']
            )

        fig = plt.figure(figsize=(5.8, 4))
        ax = plt.subplot(111, projection=rotated_pole)

        ax.set_xlim(left + border, right - border)
        ax.set_ylim(bottom + border, top - border)

        # add coastlines and countries
        ax.coastlines(resolution='10m', color=edgecolor, linewidth=0.5)
        lines = cfeature.NaturalEarthFeature(
            category='cultural',
            name='admin_0_boundary_lines_land',
            scale='10m',
        )
        ax.add_feature(lines, edgecolor=edgecolor, facecolor='none', linewidth=0.5)
        #lines = cfeature.NaturalEarthFeature(
        #    category='cultural',
        #    name='admin_1_states_provinces_lines',
        #    scale='10m',
        #)
        #ax.add_feature(lines, edgecolor=edgecolor, facecolor='none', linewidth=0.5)

    else:
        img.remove()

    # show image
    if title is not None:
        ax.set_title(title)

    img = ax.imshow(field,
        extent=(left, right, bottom, top),
        vmin=vmin, vmax=vmax, cmap=cmap
    )

    if make_new:
        cb = plt.colorbar(img, extend='both')
        cb.set_label(label)

        plt.tight_layout()
        plt.show()


    return ax, img


if __name__ == '__main__':


    make_map('config/Sentinel-7-XCO2.txt', 'berlin', 0, 1,
            add_labels=True, add_domain=True, do_overlap=False, nrow=None,
            color='black', add_pixels=True
    )
    plt.savefig('../figures/Sentinel-7-berlin.png')
    plt.savefig('../figures/Sentinel-7-berlin.pdf')
    plt.close()

    sys.exit(0)

    orbits = create_orbits('config/Sentinel-7-XCO2.txt', 0, 1, nrow=2)
    orbits += create_orbits('config/Sentinel-7-NO2_4km.txt', 0, 1, nrow=2)
    make_map(None, 'smartcarb', 0, 1,
            add_labels=True, add_domain=True, do_overlap=False, nrow=2,
            color='black', orbits=orbits
    )
    plt.savefig('../figures/Sentinel-7-orbit.png')
    plt.savefig('../figures/Sentinel-7-orbit.pdf')
    plt.close()

    make_map('config/Sentinel-7-XCO2.txt', 'europe', 0, 164,
            add_labels=True, add_domain=True, do_overlap=True, nrow=2,
            vmin=0, vmax=6
    )
    plt.savefig('../figures/Sentinel-7-XCO2-coverage.png')
    plt.savefig('../figures/Sentinel-7-XCO2-coverage.pdf')
    plt.close()


    make_map('config/Sentinel-7-NO2_4km.txt', 'europe', 0, 164,
            add_labels=True, add_domain=True, do_overlap=True, nrow=2,
            vmin=0, vmax=6
    )
    plt.savefig('../figures/Sentinel-7-NO2-coverage.png')
    plt.savefig('../figures/Sentinel-7-NO2-coverage.pdf')
    plt.close()















