#!/usr/bin/env python
# coding: utf-8

from datetime import datetime, timedelta

import osr, ogr
import numpy as np
import scipy
import scipy.ndimage as ndi

def iter_times(start, end, step=timedelta(hours=1)):
    current = start
    while current <= end:
        yield current
        current = current + step


def create_rotpole_reference(pollon, pollat):
    """\
    Return spatial reference for rotated pole coordinates.

    pollon: longitude of north pole
    pollat: latitude of north pole
    """
    if pollon is None or pollat is None:
        ref = osr.SpatialReference()
        ref.ImportFromEPSG(4326)     # WSG-84

    else:
        o_lon_p = 0.0
        o_lat_p = pollat
        lon_0 = 180 + pollon

        string = (
            'PROJCS["rotated_pole projection",EXTENSION["PROJ4","+proj=ob_tran +o_proj=longlat'
            ' +o_lon_p=%.2f +o_lat_p=%.2f +lon_0=%.2f +a=1 +to_meter=0.0174532925199"]]'
        ) % (o_lon_p, o_lat_p, lon_0)

        ref = osr.SpatialReference()
        ref.ImportFromEPSG(4326)     # WSG-84
        ref.ImportFromWkt(string)    # rotpole transformation

    return ref



def epsg2rotpole(epsg, pollon, pollat, inverse=False):
    """\
    Create transformation from EPSG to rotpole described
    by pollon, pollat.
    """
    ref0 = osr.SpatialReference()
    ref0.ImportFromEPSG(epsg)

    ref1 = create_rotpole_reference(pollon, pollat)

    if inverse:
        return osr.CoordinateTransformation(ref1, ref0)
    else:
        return osr.CoordinateTransformation(ref0, ref1)


def wgs84_to_rotpole(lon, lat, pollon, pollat):
    """\
    Convert lon, lat to rlon, rlat.
    """
    trans = epsg2rotpole(4326, pollon, pollat)
    point = ogr.Geometry(ogr.wkbPoint)
    point.AddPoint(float(lon), float(lat))
    point.Transform(trans)
    return point.GetX(), point.GetY()



def spatial_binning(a):
    kernel = 0.25 * np.ones((1,2,2))
    return ndi.convolve(a, kernel)[:,::2,::2]


