#!/usr/bin/env python
# coding: utf-8

from datetime import datetime, timedelta

from collections import defaultdict
import os
import time

import cartopy.crs as ccrs
import netCDF4
import numpy as np
import shapely

from .. import orbit
from .. import config as cfg


r_earth = 6371.0
np.seterr(divide='ignore', invalid='ignore')

def iter_pixel_area(pixel, x0=cfg.startlon, y0=cfg.startlat,
        dx=cfg.dlon, dy=cfg.dlat, ie=cfg.ie, je=cfg.je):
    """\
    Iterate over indices of grid cells and yield fraction of overlap
    of pixel and cell.
    """

    if not pixel.intersects(cfg.DOMAIN):
        raise StopIteration

    dx2, dy2 = 0.5 * dx, 0.5 * dy

    # polygon bounds
    minx, miny, maxx, maxy = pixel.bounds

    # round up and down
    minx = np.floor(minx / dx) * dx
    miny = np.floor(miny / dy) * dy
    maxx = np.ceil(maxx / dx) * dx + 0.5 * dx
    maxy = np.ceil(maxy / dy) * dy + 0.5 * dy

    # start point for pixel
    i0 = int(round((minx - x0) / dx))
    j0 = int(round((miny - y0) / dy))

    # iter over grid cells
    for i,x in enumerate(np.arange(minx,maxx,dx), i0):
        for j,y in enumerate(np.arange(miny,maxy,dy), j0):

            if 0 <= i < ie and 0 <= j < je:
                grid_cell = shapely.geometry.Polygon([
                    [x-dx2, y-dy2],
                    [x-dx2, y+dy2],
                    [x+dx2, y+dy2],
                    [x+dx2, y-dy2]
                ])

                # calculate fraction of intersection and yield
                if grid_cell.intersects(pixel):
                    weight = grid_cell.intersection(pixel).area / grid_cell.area
                    yield j, i, weight




def create_pixel(lon, lat, pollon=cfg.pollon, pollat=cfg.pollat):
    """\
    Create pixel polygon in rotpole coords from corner longitude
    and latitude.
    """
    inp = ccrs.PlateCarree()
    out = ccrs.RotatedPole(
        pole_latitude=pollat,
        pole_longitude=pollon
    )
    coords = out.transform_points(inp, lon, lat)
    pixel = shapely.geometry.Polygon(coords)

    return pixel



def create_satobs(fields, lon, lat):
    """\
    Create satellite observations from orbit filename.

    fields   :: COSMO fields as list of nc variables
    lon, lat :: longitude and latitude of satellite pixel corners
    """
    shape = lon.shape[:2]
    values = np.zeros((len(fields), shape[0], shape[1]))
    weights = np.zeros(shape)

    did_load_data = False

    for n,m in np.ndindex(shape):
        pixel = create_pixel(lon[n,m], lat[n,m])
        for j,i,weight in iter_pixel_area(pixel):

            if not did_load_data:
                fields = np.array([field[0] for field in fields])
                did_load_data = True

            values[:,n,m] += fields[:,j,i] * weight
            weights[n,m] += weight

    values /= weights[np.newaxis]
    values = np.ma.array(values, mask=np.isnan(values))

    return values



def read_orbit(orbit_filename):
    with netCDF4.Dataset(orbit_filename) as nc:
        time = nc.variables['time'][...]
        lon0 = nc.variables['longitude'][...]
        lat0 = nc.variables['latitude'][...]
        lonc = nc.variables['longitude_corners'][...]
        latc = nc.variables['latitude_corners'][...]
        sza =  nc.variables['solar_zenith_angle'][...]

    return time, lon0, lat0, lonc, latc, sza


def date2orbit(date, instrument=None):
    """\
    """
    cycle_duration, cycle_length = {
                       #  days, orbits in cycle
            'Sentinel_5': (29,  412),
            'Sentinel_7': (11,  164)
    }[instrument]

    days = (date - datetime(2015,1,1)).total_seconds() / 86400.0
    return round(days / cycle_duration * cycle_length)



def save_satobs(filename, instrument, utc_time, orbit_data=None, values=None,
        lon_center=None, lat_center=None, lon_corners=None, lat_corners=None,
        uwind=None, vwind=None):
    """\
    Write satellite orbit to file.

    If orbit_data is None lon/lat centers and corners need to be provided.
    Arrays in orbit_data need to have shape (nobs, norw, (ncorner)).
    """
    has_winds = uwind is not None and vwind is not None


    if orbit_data is None:
        nobs, nrow, ncorner = lon_corners.shape
    else:
        nobs, nrow, ncorner = orbit_data.nobs, orbit_data.nrow, orbit_data.ncorner

    if os.path.exists(filename):
        os.remove(filename)

    with netCDF4.Dataset(filename, 'w', format='NETCDF4') as nc:

        # global attributes (add input data)
        nc.setncatts({
            'DESCRIPTION':  'Synthetic Level-2 data for %s' % instrument,
            'DATAORIGIN':   'Tracer fields from COSMO-GHG simulations; Orbits from SRON orbit simulator',
            'CREATOR':      'Gerrit Kuhlmann',
            'EMAIL':        'gerrit.kuhlmann@empa.ch',
            'AFFILIATION':  'Empa Duebendorf, Switzerland',
            'VERSION':      '0.1',
            'DATE CREATED': time.ctime(time.time()),
            'STUDY':        'ESA project - Smartcarb',
            'utc_time':     str(datetime(2015,1,1) + timedelta(seconds=np.mean(utc_time))),
        })

        if orbit_data is not None:
            nc.createVariable('input', 'c').setncatts({
                # satellite
                'time_for_one_revolution_in_hours': orbit_data.trev * 24.0,
                'satellite_height_in_km': orbit_data.altitude - r_earth,
                'local_solar_time_at_the_northward_equator_pass_in_hours': orbit_data.t_node,
                'orbital_inclination_angle_in_degrees': np.rad2deg(orbit_data.inclination),

                # revolution
                'date': '%04d-%02d-%02d' % tuple(orbit_data.date),
                'longitude_at_daytime_equator_pass_in_degrees': np.rad2deg(orbit_data.longitude_eq),

                # scanlines
                'orbit_phase_at_start_of_scanning': orbit_data.phase_start,
                'orbit_phase_at_end_of_scanning': orbit_data.phase_end,
                'time_per_scan_in_seconds': orbit_data.tsamp,

                # swath
                'number_of_detector_rows': orbit_data.nrow,
                'polynomial_parameters_for_alpha': ', '.join(str(v) for v in orbit_data.polyterms_alpha),
                'polynomial_parameters_for_beta': ', '.join(str(v) for v in orbit_data.polyterms_beta),
            })


        # create dimensions
        nobs = nc.createDimension('nobs', nobs)
        nrows = nc.createDimension('nrows', nrow)
        ncorner = nc.createDimension('ncorners', ncorner)

        if has_winds:
            nlevels = nc.createDimension('nlevels', 60)

        # create variables
        nc_lat = nc.createVariable('latitude', 'f4', ('nobs', 'nrows'))
        nc_lon = nc.createVariable('longitude', 'f4', ('nobs', 'nrows'))
        nc_lat_corner = nc.createVariable('latitude_corners', 'f4', ('nobs', 'nrows', 'ncorners'))
        nc_lon_corner = nc.createVariable('longitude_corners', 'f4', ('nobs', 'nrows', 'ncorners'))

        if orbit_data is not None:
            nc_times = nc.createVariable('time', 'f8', ('nobs','nrows'))
            nc_sza = nc.createVariable('solar_zenith_angle', 'f4', ('nobs', 'nrows'))
            nc_vza = nc.createVariable('viewing_zenith_angle', 'f4', ('nobs', 'nrows'))
            nc_vaz = nc.createVariable('viewing_azimuth_angle', 'f4', ('nobs', 'nrows'))
            nc_saz = nc.createVariable('solar_azimuth_angle', 'f4', ('nobs', 'nrows'))

        if has_winds:
            nc_uwind = nc.createVariable('u-wind', 'f4', ('nobs', 'nrows', 'nlevels'))
            nc_vwind = nc.createVariable('v-wind', 'f4', ('nobs', 'nrows', 'nlevels'))

        # Assign an attribute to the dataset
        nc_lat.standard_name = "latitude"
        nc_lat.long_name = "pixel center latitude"
        nc_lat.units = "degrees north"

        nc_lon.standard_name = "longitude"
        nc_lon.long_name = "pixel center longitude"
        nc_lon.units = "degrees east"

        nc_lat_corner.standard_name = "latitude"
        nc_lat_corner.long_name = "pixel corners latitude"
        nc_lat_corner.units = "degrees north"

        nc_lon_corner.standard_name = "longitude"
        nc_lon_corner.long_name = "pixel corners longitude"
        nc_lon_corner.units = "degrees east"

        if orbit_data is not None:
            nc_times.standard_name = "time_of_measurement"
            nc_times.long_name = "time of measurement since 2015-01-01 00:00:00"
            nc_times.units = "seconds since 2015-01-01 00:00:00"

            nc_sza.standard_name = "solar_zenith_angle"
            nc_sza.long_name = "solar zenith angle"
            nc_sza.units = "degree"

            nc_vza.standard_name = "viewing_zenith_angle"
            nc_vza.long_name = "viewing zenith angle"
            nc_vza.units = "degree"

            nc_vaz.standard_name = "viewing_azimuth_angle"
            nc_vaz.long_name = "viewing azimuth angle"
            nc_vaz.units = "degree"

            nc_saz.standard_name = "solar_azimuth_angle"
            nc_saz.long_name = "solar azimuth angle"
            nc_saz.units = "degree"


        if has_winds:
            nc_uwind.standard_name = 'grid_eastward_wind'
            nc_uwind.long_name = 'U-component of wind'
            nc_uwind.units = 'm s-1'

            nc_vwind.standard_name = 'grid_northward_wind'
            nc_vwind.long_name = 'V-component of wind'
            nc_vwind.units = 'm s-1'

        # fill variables
        nc_lat[:,:] = lat_center
        nc_lon[:,:] = lon_center
        nc_lat_corner[:,:,:] = lat_corners
        nc_lon_corner[:,:,:] = lon_corners

        if orbit_data is not None:
            pass

        if has_winds:
            nc_uwind[:,:,:] = uwind
            nc_vwind[:,:,:] = vwind

        for key in values.keys():
            v = nc.createVariable(key, 'f4', ('nobs', 'nrows'))
            v[:,:] = values[key]

            v.standard_name = cfg.tracer2long_name[key]
            v.long_name = cfg.tracer2long_name[key]

            if key == 'CLCT':
                v.units = '1'
            elif key.startswith('XCO2'):
                v.units = 'ppm'
            else:
                v.units = 'molecules cm-2'


def calculate_sza():
    pass




def main(instrument, orbit_number, names, path_orbits, path_2d_fields,
        path_level2_data, lon_eq=0.0, utc=None, online_orbit=True):
    """\
    Create Level-2 for `orbit_number` of satellite `instrument`
    (e.g. Sentinel_7_CO2). Variables are taken from COSMO-2D fields.

    TODO:
        - add support for U- and V-wind from 3D fields
        - add MODIS surface reflectance
        - add cloud properties (???)
        - add errors for different scenarios
    """
    if online_orbit:
        orbit_data = orbit.compute_orbit(instrument, orbit_number,
                long_eq=lon_eq, do_reshape=True, write_to_file=False
        )
        time = orbit_data.times
        lon0 = orbit_data.lon_center
        lat0 = orbit_data.lat_center
        lonc = orbit_data.lon_corner
        latc = orbit_data.lat_corner
        sza = None

        cycle_duration = 0.0
        cycle_number = 0

    else:
        orbit_data = None

        # get cycle and orbit number
        if instrument == 'Sentinel_5':
            cycle_duration = 29.0 # days
            cycle_number, number  = divmod(orbit_number, 412)
        else:
            cycle_duration = 11.0 # days
            cycle_number, number  = divmod(orbit_number, 164)

        # get lon, lat, utc_time from orbit
        orbit_filename = os.path.join(
                path_orbits, instrument,
                '%s_o%03d_l000_pixel.nc' % (instrument, number)
        )
        time, lon0, lat0, lonc, latc, sza = read_orbit(orbit_filename)

        # TODO: update time and SZA

        # shift equator passing longitude, e.g. 0,90,180 or 270
        # for four satellites
        lon0 = (lon0 + lon_eq + 180.0) % 360 - 180.0
        lonc = (lonc + lon_eq + 180.0) % 360 - 180.0

        if np.any((lonc < -180.0) | (lonc > 180.0)):
            raise ValueError


    # check if any latc, lonc is in domain
    if not (lon0.min() < 19.8 and lon0.max() > 7.5):
        print('%s orbit %d with lon %d is outside domain' % (instrument, orbit_number, lon_eq))
        return

    # use closest COSMO time or provided time
    if utc is None:

        # seconds since 2015-01-01 00:00:00"
        utc = datetime(2015,1,1) + timedelta(seconds=np.mean(time))

        # add cycles
        utc += timedelta(days=cycle_number*cycle_duration)

        # round to next hour (TODO: use linear interpolation?)
        utc = datetime(utc.year, utc.month, utc.day) + timedelta(hours=round(utc.hour + utc.minute / 60.0))

    # open 2d fields
    cosmo_filename = os.path.join(path_2d_fields, utc.strftime('cosmo_2d_%Y%m%d%H.nc'))

    # create satellite observations
    try:
        with netCDF4.Dataset(cosmo_filename) as nc:
            if names is None:
                if instrument == 'Sentinel_7_CO2':
                    names = sorted([
                        key for key in nc.variables.keys()
                        if key.startswith('XCO2') or key == 'CLCT'
                    ])
                else:
                    names = sorted([
                        key for key in nc.variables.keys()
                        if key.endswith('_column') or key == 'CLCT' # TODO
                    ])

            variables = [nc.variables[name] for name in names]
            satobs = create_satobs(variables, lonc, latc)

    except OSError:
        print('failed to open "%s"' % cosmo_filename)
        return

    if np.all(satobs.mask):
        print('%s orbit %d with lon %d is outside domain' % (instrument, orbit_number, lon_eq))
        return

    print('%s orbit %d with lon %d is inside domain' % (instrument, orbit_number, lon_eq))

    satobs = dict((name, satobs[i]) for i,name in enumerate(names))

    # output filename
    output_filename = os.path.join(
            path_level2_data,
            '%s_%s_o%04d_l%03d_data.nc' % (instrument, utc.strftime('%Y%m%d%H'), orbit_number, lon_eq)
    )
    save_satobs(
            output_filename, instrument, time, orbit_data, satobs,
            lon0, lat0, lonc, latc
    )

    return utc









